--[[

  LibScrollyTextThing

  Handles making a scrollable text area (optionally editable) conveniently!


  scrolly = LibScrollyTextThing.create(frame, addon, text, flags, side, changed)
	Creates a scrollbar and text frame in <frame>, containing <text>,
	with behavior according to <flags>.  Will call
		changed(scrolly, error)
	whenever the text changes, if changed was provided.  side should
	be LEFT or RIGHT.  If there is an error condition, it will be passed
	as the second argument to changed.

	scrolly.background
		The background frame that text is drawn over.
	scrolly.textarea
		The foreground frame.
	scrolly.nextfield
		If present, tab will jump to this field, instead of
		inserting a tab.

	scrolly:text(new)
		Returns the current text; if new is non-nil, sets text
		to it.

	flags:
		editable:  RiftTextField, not Text, object.
		autoindent:  Whether to autoindent.  Ignored if not editable.
		wordwrap:  Whether to wordwrap.
		font: Font to use.
		fontsource: Font source to use (or Rift)
		size: Size of font to use.
		number:  Line numbers.
		html: Enable the (limited) HTML support of Rift Text objects.
			Does not work on editable objects.
		max: Maximum length.
		glow: glow effect (a table for SetEffectGlow)

]]--


Library = Library or {}
local addoninfo, stt = ...
Library.LibScrollyTextThing = stt

local printf = Library.printf.printf
local sprintf = Library.printf.sprintf

function stt.check_scrollbar(frametable)
  if not frametable or type(frametable) ~= 'table' then
    printf("check_scrollbar: Called without a valid table.")
    return
  end
  if not frametable.scrollbar then
    printf("check_scrollbar: Called without a scrollbar.")
    return
  end
  local data = frametable:text()
  local empty_line = string.match(data, '[\r\n]$')
  -- ignore HTML, since it might not matter.
  frametable.hiddenarea:SetText(data)
  local height = frametable.hiddenarea:GetHeight()
  local lines = math.ceil((height - frametable.line_slush) / frametable.line_height)
  -- the library doesn't assign any height to a trailing empty line.
  if empty_line then
    lines = lines + 1
    height = height + frametable.line_height
  end
  frametable.textarea:SetHeight(math.max(height, frametable.height))
  if lines <= frametable.lines then
    frametable.scrollbar:SetEnabled(false)
    frametable.scrollbar:SetRange(0, 1)
    frametable.scrollbar:SetPosition(0)
    frametable.offset = 0
    frametable.pending_offset = nil
    frametable.textarea:SetPoint("TOPLEFT", frametable.mask, "TOPLEFT", 0, 0 - frametable.line_slush / 2)
  else
    frametable.offset = frametable.offset or 0
    frametable.scrollbar:SetEnabled(true)
    frametable.scrollbar:SetRange(0, lines - frametable.lines)
    frametable.offset = math.floor(frametable.scrollbar:GetPosition())
    if frametable.pending_offset then
      frametable.offset = frametable.offset + frametable.pending_offset
      frametable.pending_offset = nil
      -- don't try to scroll past the end (a problem if offset was 0 and
      -- we added two lines, when there was one open line already.)
      if frametable.offset > (lines - frametable.lines) then
        frametable.offset = lines - frametable.lines
      end
      frametable.scrollbar:SetPosition(frametable.offset)
    end
    frametable.textarea:SetPoint("TOPLEFT", frametable.mask, "TOPLEFT", 0, 0 - frametable.offset * frametable.line_height - frametable.line_slush / 2)
  end
  if frametable.numbers then
    for i = 1, frametable.lines do
      if frametable.offset + i <= lines then
        frametable.numbers[i]:SetText(tostring(frametable.offset + i))
        frametable.numbers[i]:SetVisible(true)
      else
        frametable.numbers[i]:SetVisible(false)
      end
    end
  end
end

function stt.do_set_text(frametable, textarea, text)
  if frametable.max then
    if #text > frametable.max then
      frametable.error = sprintf("Too long: %d/%d bytes.", #text, frametable.max)
      text = text:sub(1, frametable.max)
    end
  end
  if frametable.html then
    -- you don't even want to know
    textarea:SetText(text, false)
    return textarea:SetText(text, true)
  else
    return textarea:SetText(text)
  end
end

function stt.set_text(frametable, text)
  local old = frametable.textarea:GetText()
  if text then
    stt.do_set_text(frametable, frametable.textarea, text)
    if frametable.flags.editable then
      frametable.textarea:SetSelection()
    end
    frametable:check_scrollbar()
  end
  return old
end

function stt.key_down(frametable, key)
  local ta = frametable.textarea
  local position = ta:GetCursor()
  local text = ta:GetText()
  local before = string.sub(text, 1, position)
  local after = string.sub(text, position + 1)
  local byte = string.byte(key)
  if key == 'Tab' then
    if frametable.nextfield then
      frametable.nextfield:SetKeyFocus(true)
    else
      stt.do_set_text(frametable, ta, before .. "\t" .. after)
      position = position + 1
      if frametable.max and position >= frametable.max then
        position = frametable.max
      end
      ta:SetCursor(position)
      frametable:change()
    end
  elseif key == 'Return' then
    local insert = "\r"
    frametable.pending_offset = 1
    position = position + 1

    if frametable.flags.autoindent then
      tlb = string.match(before, "([^\n\r]*)$") or ''
      leading_space = string.match(tlb, "^(%s*)")
      if leading_space then
        insert = insert .. leading_space
	position = position + #leading_space
      end
      if string.match(tlb, " do%s*$") or string.match(tlb, " then%s*$") or string.match(tlb, "^%s*function ") or string.match(tlb, "^%s*local%s*function ") then
        insert = insert .. "  \r" .. leading_space .. 'end'
	position = position + 2
	frametable.pending_offset = 2
      end
      if string.match(tlb, "{%s*$") then
        insert = insert .. "  \r" .. leading_space .. '}'
	position = position + 2
	frametable.pending_offset = 2
      end
      if string.match(tlb, "else$") then
        insert = insert .. "  "
	position = position + 2
      end
    end
    stt.do_set_text(frametable, ta, before .. insert .. after)
    if frametable.max and position >= frametable.max then
      position = frametable.max
    end
    ta:SetCursor(position)
    frametable:change()
  else
    if frametable.max and #text > frametable.max then
      -- truncate
      stt.do_set_text(frametable, ta, text)
      text = ta:GetText()
      position = #text
      ta:SetCursor(position)
    end
  end
end

function stt.change(frametable)
  if frametable.changed then
    frametable:changed(frametable.error)
    -- only report any given error once
    frametable.error = nil
  end
  frametable:check_scrollbar()
end

-- allow for height changes
function stt.size_change(frametable, args)
  local h = frametable.frame:GetHeight()
  frametable.height = h
  frametable.scrollbar:SetHeight(h)
  frametable.background:SetHeight(h)
  frametable.mask:SetHeight(h)
  frametable:check_scrollbar()
end

function stt.create(frame, addon, text, flags, side, changed)
  local frametable = { u = {}, frame = frame, addon = addon, flags = flags, changed = changed }
  local w = frame:GetWidth()
  local h = frame:GetHeight()
  local scroll_side = 'TOP' .. side
  local other_side
  local away = 0
  flags = flags or {}

  if w < 40 or h < 40 then
    printf("Provided frame not large enough.")
    return
  end
  if side == 'LEFT' then
    other_side = 'TOPRIGHT'
    away = 1
  elseif side == 'RIGHT' then
    other_side = 'TOPLEFT'
    away = -1
  else
    printf("Side must be LEFT or RIGHT.")
    return
  end
  frametable.scroll_side = scroll_side
  frametable.other_side = other_side
  frametable.scrollbar = UI.CreateFrame('RiftScrollbar', addon, frame)
  frametable.scrollbar:SetPoint('TOP' .. side, frame, 'TOP' .. side)
  frametable.scrollbar:SetHeight(h)
  frame:EventAttach(Event.UI.Layout.Size, function(event) stt.size_change(frametable, event) end, "handle size changes")
  local sbw = frametable.scrollbar:GetWidth()
  w = w - sbw - 1
  frame:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function() frametable.scrollbar:Nudge(3) end, "wheel back")
  frame:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function() frametable.scrollbar:Nudge(-3) end, "wheel forward")
  frametable.scrollbar:EventAttach(Event.UI.Scrollbar.Change, function() frametable:check_scrollbar() end, "scroll bar")


  frametable.height = h
  frametable.width = w
  -- non-editable fields can have HTML enabled
  frametable.html = flags.html and not flags.editable
  -- support maximum length
  frametable.max = flags.max

  frametable.changed = changed 
  frametable.offset = 0
  frametable.background = UI.CreateFrame('Frame', addon, frame)

  frametable.mask = UI.CreateFrame('Mask', addon, frametable.background)

  if flags.number then
    frametable.number_frame = UI.CreateFrame('Frame', addon, frametable.background)
    frametable.number_frame:SetWidth(30)
    frametable.number_frame:SetHeight(h)
    w = w - 32
  end
  frametable.background:SetPoint(scroll_side, frametable.scrollbar, other_side, away, 0)
  frametable.background:SetWidth(w)
  frametable.background:SetHeight(h)

  -- add two pixel border for the textarea
  h = h - 4
  w = w - 4

  frametable.mask:SetPoint("TOPLEFT", frametable.background, "TOPLEFT", 2, 2)
  frametable.mask:SetWidth(w)
  frametable.mask:SetHeight(h)

  if flags.editable then
    frametable.textarea = UI.CreateFrame('RiftTextfield', addon, frametable.mask)
    frametable.textarea:EventAttach(Event.UI.Input.Key.Down, function(field, event, key) frametable:key_down(key) end, "key down")
    frametable.textarea:EventAttach(Event.UI.Textfield.Change, function() frametable:change() end, "textfield changed")
  else
    frametable.textarea = UI.CreateFrame('Text', addon, frametable.mask)
  end
  frametable.hiddenarea = UI.CreateFrame('Text', addon, frametable.mask)
  frametable.hiddenarea:SetVisible(false)
  frametable.textarea:SetPoint("TOPLEFT", frametable.mask, "TOPLEFT", 0, 0)
  frametable.textarea:SetWidth(w)
  stt.do_set_text(frametable, frametable.textarea, text or '')
  if flags.size and not flags.editable then
    frametable.textarea:SetFontSize(flags.size)
    frametable.hiddenarea:SetFontSize(flags.size)
  end
  if flags.glow and not flags.editable then
    frametable.textarea:SetEffectGlow(flags.glow)
  end
  if flags.font and not flags.editable then
    frametable.textarea:SetFont(flags.fontsource or 'Rift', flags.font)
    frametable.hiddenarea:SetFont(flags.fontsource or 'Rift', flags.font)
  end
  if flags.wordwrap and not flags.editable then
    frametable.textarea:SetWordwrap(flags.wordwrap)
    frametable.hiddenarea:SetWordwrap(flags.wordwrap)
    -- and we need to cap the width for that to work.
    frametable.hiddenarea:SetWidth(w)
  end

  if not flags.editable then

    local five_lines = 'a\nb\nc\nd\ne'
    frametable.hiddenarea:SetText(five_lines)
    local height5 = frametable.hiddenarea:GetHeight()
    frametable.hiddenarea:SetText(five_lines .. '\n' .. five_lines)
    local height10 = frametable.hiddenarea:GetHeight()
    frametable.line_height = (height10 - height5) / 5
    if frametable.line_height < 5 then
      printf("Warning: Line height was unreasonably small, bumping it a little.")
      frametable.line_height = 6
    end
    frametable.line_slush = (height5 - (frametable.line_height * 5))
    frametable.lines = math.floor(h / frametable.line_height)
  else
    frametable.line_height = 12.6
    frametable.line_slush = 5
    frametable.lines = math.floor(h / frametable.line_height)
  end
  if flags.number then
    frametable.numbers = {}
    local number_count = math.ceil(h / frametable.line_height) + 1
    for i = 1, number_count do
      frametable.numbers[i] = UI.CreateFrame('Text', addon, frametable.number_frame)
      frametable.numbers[i]:SetPoint("TOPRIGHT", frametable.number_frame, "TOPRIGHT", 0, frametable.line_height * (i - 1))
      frametable.numbers[i]:SetFontColor(0.6, 0.6, 0.45)
    end
    frametable.number_frame:SetPoint("TOPRIGHT", frametable.background, "TOPLEFT", 0, 0)
  end

  frametable.check_scrollbar = stt.check_scrollbar
  frametable.text = stt.set_text
  frametable.SetKeyFocus = function(self, state) frametable.textarea:SetKeyFocus(state) end
  frametable.key_down = stt.key_down
  frametable.key_type = stt.key_type
  frametable.change = stt.change
  frametable:check_scrollbar()
  return frametable
end
